struct lease_details * delete_lease_node (struct lease_details *, struct lease_details *);
struct lease_details * read_leases_file (char *);
void write_leases_file (char *, struct lease_details *);
void check_leases_for_expiry (void);
void get_next_address (struct in6_addr *);
int get_pref_address (struct addr_details *, struct lease_details *, struct lease_details *, struct config_head *, struct sockaddr_in6, struct OPTIONS *);
int get_available_address (struct addr_details *, struct lease_details *, struct lease_details *, struct config_head *,	struct sockaddr_in6);
int check_if_addr_in_lease_structure (struct in6_addr, struct lease_details *);
void add_node_to_partial_lease_structure (struct DHCP_MESSAGE *, struct addr_details *);
int check_for_match (struct DHCP_MESSAGE *, struct lease_details *);
int check_for_duid_match (struct DHCP_MESSAGE *, struct lease_details *);
int move_node_from_partial_to_lease (struct DHCP_MESSAGE *);
int move_node_from_lease_to_partial (struct DHCP_MESSAGE *);
void remove_expired_nodes_from_partial_lease (void);
void del_node_from_lease (struct DHCP_MESSAGE *, struct lease_details *);
int check_renewal_policy (struct DHCP_MESSAGE *, struct lease_details *);
